__all__ = ["FrameFEM", "Material", "Section", "BeamSection",
           "Support", "Load", "PointLoad", "LineLoad", "LoadCase",
           "FEMNode", "Element"]
#from .framefem import FrameFEM, BeamSection, PointLoad, LineLoad, Element, Section, Material
from .framefem import *
name = "framefem"

