# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Mon Jan  4 12:45:41 2021

Constants from EN 1995

@author: Viktor Haimi
"""
# Partial safety factors taken from EN 1995-1-1 § 2, Table 2.3
gammaM_st = 1.3
gammaM_glt = 1.25
gammaM_lvl = 1.2
gammaM_other = 1.3

# SFS-EN 1995-1-1 § 6.1.6
k_m = 0.7
