#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

# @Filename : __init__.py
# @Date : 2022
# @Project: metku
# @AUTHOR : Jaakko Huusko
